/*
 * NokiaDisplay.c
 * 
 * Демонстрация работы с дисплеями Nokia (на базе чипов PCD8544)
 * 
 * Помещённый здесь код является свободным. Т.е. допускается его свободное использование для любых целей, включая коммерческие, при условии указания ссылки на автора (Погребняк Дмитрий, http://aterlux.ru/).
 * Code included here is free to use. It can be used for free for any purpose, including commercial use, as long as link to author (Pogrebnyak Dmitry, http://aterlux.ru/) is included.
 */ 


#include <avr/io.h>
#include <avr/pgmspace.h>
#include "PCD8544.h"
#include <util/delay.h>
#include "text.h"
#include "tools.h"

// Внеэкранный буфер. Возьмём ширину 64 - для круглости
uint8_t buf[64 * 6];

// Очистка буфера
void buf_clear() {
  uint16_t * pb = (void*)&buf[0];
  for (uint8_t i = 0; i < 192; i++)
    *(pb++) = 0;
}

// Рисование линии.
void buf_line(int8_t x1, int8_t y1, int8_t x2, int8_t y2) {
  uint8_t dx = (x2 < x1) ? (x1 - x2) : (x2 - x1);
  uint8_t dy = (y2 < y1) ? (y1 - y2) : (y2 - y1);
  if (dx > dy) { // Линия ближе к горизонтальной
    if (x2 < x1) { // Рисуем всё время слева направо
      int8_t t = x2; x2 = x1; x1 = t;
      t = y2; y2 = y1; y1 = t;
    }
    int8_t page = y1 >> 3;
    uint8_t page_mask = 1 << (y1 & 7);
    uint8_t yinc = (y2 > y1) ? 1 : 0;
    uint16_t err = dx >> 1;
    if (x2 > 63)
      x2 = 63;
    while (x1 <= x2) {
      if ((x1 >= 0) && (page >= 0) && (page < 6)) {
        buf[page * 64 + x1] |= page_mask;
      }
      err += dy;
      if (err >= dx) {
        err -= dx;
        if (yinc) {
          page_mask <<= 1;
          if (!page_mask) {
            page++;
            if (page >= 6)
              return;
            page_mask = 0x01;
          }
        } else {
          page_mask >>= 1;
          if (!page_mask) {
            if (page <= 0)
              return;
            page--;
            page_mask = 0x80;
          }
        }
      }
      x1++;
    }
  } else { // Линия ближе к вертикальной
    if (y2 < y1) { // Рисуем всё время сверху вниз
      int8_t t = x2; x2 = x1; x1 = t;
      t = y2; y2 = y1; y1 = t;
    }
    uint8_t page = y1 >> 3;
    uint8_t page_mask = 1 << (y1 & 7);
    int8_t xvector = (x2 > x1) ? 1 : -1;
    uint16_t err = dy >> 1;
    if (y2 > 47)
      y2 = 47;
    while (y1 <= y2) {
      if ((x1 >= 0) && (x1 < 64) && (page >= 0)) {
        buf[page * 64 + x1] |= page_mask;
      }
      err += dx;
      if (err >= dy) {
        err -= dy;
        x1 += xvector;
      }
      page_mask <<= 1;
      if (!page_mask) {
        page++;
        page_mask = 1;
      }
      y1++; 
    }
  }
}

typedef struct {
  int8_t x, y, z;
} Point3D;

typedef struct {
  int8_t x, y;
  uint8_t isVisible;
} Point3Dv;

typedef struct {
  uint8_t p1, p2;
} Line;

typedef struct {
  uint8_t cnt;
  uint8_t points[5];
} Surface;

typedef struct {
  int8_t x, y;
  uint8_t z;
  int8_t sx, sy;
} Star;

const Point3D points[] PROGMEM = {
  {-50, -50, -50},
  {50, -50, -50},
  {-50, 50, -50},
  {50, 50, -50},
  {-50, -50, 50},
  {-50, 50, 50},
  {50, 50, 50},
  {50, 0, 50},
  {50, -50, 0},
  {0, -50, 50},
};

const Line lines[] PROGMEM = {
  {0, 1}, {1, 3}, {3, 2}, {2, 0},
  {0, 4}, {4, 5}, {5, 2}, {5, 6}, {6, 3},
  {7, 8}, {8, 9}, {9, 7}, 
  {4, 9}, {1, 8}, {6, 7}
};


const Surface surfaces[] PROGMEM = {
  {4, {0, 2, 3, 1}},
  {4, {0, 4, 5, 2}},
  {4, {5, 6, 3, 2}},
  {5, {4, 0, 1, 8, 9}},
  {5, {6, 5, 4, 9, 7}},
  {5, {1, 3, 6, 7, 8}},
  {3, {9, 8, 7}},
};

const int16_t sintable[] PROGMEM = {
  0, 6, 13, 19, 25, 31, 38, 44, 50, 56, 62, 68, 74, 80, 86, 92, 98,
  104, 109, 115, 121, 126, 132, 137, 142, 147, 152, 157, 162, 167, 172, 177, 181,
  185, 190, 194, 198, 202, 206, 209, 213, 216, 220, 223, 226, 229, 231, 234, 237,
  239, 241, 243, 245, 247, 248, 250, 251, 252, 253, 254, 255, 255, 256, 256, 256
};

inline int16_t tsin(uint8_t period) {
  if (period >= 0x80) {
    return (period > 0xC0) ? -pgm_read_word(&sintable[0x100 - period]) : -pgm_read_word(&sintable[period - 0x80]);
  } else {
    return (period > 0x40) ? pgm_read_word(&sintable[0x80 - period]) : pgm_read_word(&sintable[period]);
  }
}

inline int16_t tcos(uint8_t period) {
  if (period >= 0x80) {
    return (period > 0xC0) ? pgm_read_word(&sintable[period - 0xC0]) : -pgm_read_word(&sintable[0xC0 - period]);
  } else {
    return (period > 0x40) ? -pgm_read_word(&sintable[period - 0x40]) : pgm_read_word(&sintable[0x40 - period]);
  }
}

Point3Dv vpoints[arraysize(points)];

inline void rotate_all_points(uint8_t xAngle, uint8_t yAngle, uint8_t x, uint8_t y, uint8_t distAdd) {
  for (uint8_t i = 0; i < arraysize(points); i++) {
    int16_t ax = (int8_t)pgm_read_byte(&points[i].x);
    int16_t ay = (int8_t)pgm_read_byte(&points[i].y);
    int16_t az = (int8_t)pgm_read_byte(&points[i].z);
    int16_t s, c;
    s = tsin(xAngle);
    c = tcos(xAngle);
    int16_t t = (ay * c - az * s) / 256;
    az = (az * c + ay * s) / 256;
    ay = t;
    s = tsin(yAngle);
    c = tcos(yAngle);
    t = (ax * c - az * s) / 256;
    az = (az * c + ax * s) / 256;
    ax = t;
    az = az + 200 + distAdd;
    vpoints[i].x = ax * 64 / az + x; // Изображение анизоморфное, поэтому множитель разный для X и Y
    vpoints[i].y = ay * 54 / az + y;
    vpoints[i].isVisible = 0;
  }
}

inline void find_visible_points() {
  for (uint8_t i = 0; i < arraysize(surfaces); i++) {
    uint8_t p1 = pgm_read_byte(&surfaces[i].points[0]);
    uint8_t p2 = pgm_read_byte(&surfaces[i].points[1]);
    uint8_t p3 = pgm_read_byte(&surfaces[i].points[2]);
    int16_t dx1 = (int16_t)vpoints[p1].x - vpoints[p2].x;
    int16_t dy1 = (int16_t)vpoints[p1].y - vpoints[p2].y;
    int16_t dx2 = (int16_t)vpoints[p3].x - vpoints[p2].x;
    int16_t dy2 = (int16_t)vpoints[p3].y - vpoints[p2].y;

    int16_t v = (int16_t)(dx1 * dy2) - (int16_t)(dy1 * dx2);
    if (v > 0) {
      vpoints[p1].isVisible = 1;
      vpoints[p2].isVisible = 1;
      vpoints[p3].isVisible = 1;
      uint8_t pcnt = pgm_read_byte(&surfaces[i].cnt);
      if (pcnt >= 4) 
        vpoints[pgm_read_byte(&surfaces[i].points[3])].isVisible = 1;
      if (pcnt >= 5)
        vpoints[pgm_read_byte(&surfaces[i].points[4])].isVisible = 1;

    }
  }
}

inline void draw_all_lines() {
  for (uint8_t i = 0; i < arraysize(lines); i++) {
    uint8_t p1 = pgm_read_byte(&lines[i].p1);
    uint8_t p2 = pgm_read_byte(&lines[i].p2);
    if (vpoints[p1].isVisible && vpoints[p2].isVisible)
      buf_line(vpoints[p1].x, vpoints[p1].y, vpoints[p2].x, vpoints[p2].y);
  } 
}

void delay(uint16_t ms) {
  while (ms--) {
    _delay_ms(1);
  }
}

PROGMEM char const pstr_normal_rus[] = "Обычный";
PROGMEM char const pstr_normal_eng[] = "Normal";
PROGMEM char const pstr_narrow_rus[] = "Узкий";
PROGMEM char const pstr_narrow_eng[] = "Narrow";
PROGMEM char const pstr_bold_rus[] = "Жирный";
PROGMEM char const pstr_bold_eng[] = "Bold";
PROGMEM char const pstr_handwrite_rus[] = "Рукописный";
PROGMEM char const pstr_handwrite_eng[] = "Handwrite";

PROGMEM PGM_VOID_P const p_fontnames[] = {
  pstr_normal_rus, pstr_normal_eng,
  pstr_narrow_rus, pstr_narrow_eng,
  pstr_bold_rus, pstr_bold_eng,
  pstr_handwrite_rus, pstr_handwrite_eng
};


int main(void)
{
  PORTD = 0;
  DDRD = 0b00011111;
  
  
  display_init();
  
  
  uint8_t xAngle1 = 123;
  uint8_t yAngle1 = 32;
  uint16_t distPeriod1 = 0;
  while(1) {
    display_clear();
    text_set_font(FONT_BOLD);
    text_print(0, 0, PSTR("PCD8544"), 0);
    text_set_font(FONT_NORMAL);
    text_print(2, 0, PSTR("Автор:"), 0);
    text_print(3, 0, PSTR("Погребняк Д.Э."), 0);
    text_set_font(FONT_HANDWRITE);
    text_print(5, 0, PSTR("г Самара, 2014"), 0);
    delay(7000);
    
    for (uint8_t font = 0; font < 4; font++) {
      text_set_font(font);
      display_clear();
      text_print(0, 0, (PGM_VOID_P)(uint_fast32_t)pgm_read_word(&p_fontnames[font * 2]), 0);
      text_print(1, 0, (PGM_VOID_P)(uint_fast32_t)pgm_read_word(&p_fontnames[font * 2 + 1]), 0);
      text_print(3, 0, PSTR("Всем привет!"), 0);
      text_print(4, 0, PSTR("Hello, World!"), 0);
      text_print(5, 0, PSTR("01234356789?!:;-+"), 0);
      delay(5000);
      display_clear();
      uint8_t y = 0, x = 0; 
      for (uint8_t i = 23; i < 194; i++) {
        uint8_t w = text_char_width(i);
        if (w > DISPLAY_WIDTH - x) {
          y++;
          x = 0;
          if (y >= 6) {
            _delay_ms(2000);
            display_clear();
            y = 0;
          }
        }
        text_char(y, x, i, 0);
        delay(10);
        x += w;
      }
      delay(3000);
    } 
    display_clear();  
    
    for (uint16_t f = 0; f < 500; f++) {
      xAngle1 += 3;
      yAngle1 += 1;
      distPeriod1 += 439;
      rotate_all_points(xAngle1, yAngle1, 32, 24, ((tsin(distPeriod1 >> 8) + 280) * 100) >> 8);
      find_visible_points();
      buf_clear();
      draw_all_lines();
      display_buffer(0, 10, &buf[0], 6, 64);
      delay(70);
        //TODO:: Please write your application code 
    }     
  }
}